﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis {
  public enum TerminationCriterium { OnlyByTime, OnlyByEvaluations, OnlyByTarget, ByTargetAndTime, ByTargetAndEvaluations, WhicheverHitsFirst, WhicheverHitsLast }
  /// <summary>
  /// A run in which an algorithm is executed for a certain maximum time only.
  /// </summary>
  [Item("Independent Random Restart Run", "A run in which an optimizer is repeated until either a certain target value is reached or a maximum budget is exceeded.")]
  [Creatable(CreatableAttribute.Categories.TestingAndAnalysis, Priority = 117)]
  [StorableClass]
  public sealed class IndepdentRandomRestartRun : NamedItem, IOptimizer, IStorableContent, INotifyPropertyChanged {
    public string Filename { get; set; }

    #region ItemImage
    public static new Image StaticItemImage {
      get { return VSImageLibrary.Event; }
    }
    public override Image ItemImage {
      get { return (Algorithm != null) ? Algorithm.ItemImage : VSImageLibrary.ExecutableStopped; }
    }
    #endregion

    [Storable]
    private TerminationCriterium terminationCriterium;
    public TerminationCriterium TerminationCriterium {
      get { return terminationCriterium; }
      set {
        if (terminationCriterium == value) return;
        terminationCriterium = value;
        OnPropertyChanged("TerminationCriterium");
      }
    }

    [Storable]
    private TimeSpan maximumExecutionTime;
    public TimeSpan MaximumExecutionTime {
      get { return maximumExecutionTime; }
      set {
        if (maximumExecutionTime == value) return;
        maximumExecutionTime = value;
        OnPropertyChanged("MaximumExecutionTime");
      }
    }

    [Storable]
    private int maximumEvaluations;
    public int MaximumEvaluations {
      get { return maximumEvaluations; }
      set {
        if (maximumEvaluations == value) return;
        maximumEvaluations = value;
        OnPropertyChanged("MaximumEvaluations");
      }
    }

    [Storable]
    private double targetValue;
    public double TargetValue {
      get { return targetValue; }
      set {
        if (targetValue == value) return;
        targetValue = value;
        OnPropertyChanged("TargetValue");
      }
    }

    [Storable]
    private bool maximization;
    public bool Maximization {
      get { return maximization; }
      private set {
        if (maximization == value) return;
        maximization = value;
        OnPropertyChanged("Maximization");
      }
    }

    [Storable]
    private double moveCostPerSolution;
    public double MoveCostPerSolution {
      get { return moveCostPerSolution; }
      set {
        if (moveCostPerSolution == value) return;
        moveCostPerSolution = value;
        OnPropertyChanged("MoveCostPerSolution");
      }
    }


    public ExecutionState ExecutionState {
      get { return (Algorithm != null) ? Algorithm.ExecutionState : ExecutionState.Stopped; }
    }

    private TimeSpan lastAlgorithmExecutionTime;
    [Storable]
    private TimeSpan executionTime;
    public TimeSpan ExecutionTime {
      get { return executionTime; }
      set {
        if (executionTime == value) return;
        executionTime = value;
        OnPropertyChanged("ExecutionTime");
        OnExecutionTimeChanged();
      }
    }

    private int lastAlgorithmEvaluatedSolutions;
    private int lastAlgorithmEvaluatedMoves;
    [Storable]
    private double evaluations;
    public double Evaluations {
      get { return evaluations; }
      set {
        if (evaluations == value) return;
        evaluations = value;
        OnPropertyChanged("Evaluations");
      }
    }

    [Storable]
    private double bestSoFar;
    public double BestSoFar {
      get { return bestSoFar; }
      set {
        if (bestSoFar == value) return;
        bestSoFar = value;
        OnPropertyChanged("BestSoFar");
      }
    }

    [Storable]
    private IRun currentRun;
    public IRun CurrentRun {
      get { return currentRun; }
      private set {
        if (currentRun == value) return;
        currentRun = value;
        OnPropertyChanged("CurrentRun");
      }
    }

    [Storable]
    private IAlgorithm algorithm;
    public IAlgorithm Algorithm {
      get { return algorithm; }
      set {
        if (algorithm == value) return;
        if (algorithm != null) DeregisterAlgorithmEvents();
        algorithm = value;
        if (algorithm != null) {
          RegisterAlgorithmEvents();
        }
        OnPropertyChanged("Algorithm");
        Prepare();
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
      private set {
        if (value == null) throw new ArgumentNullException();
        if (runs == value) return;
        runs = value;
        OnPropertyChanged("Runs");
      }
    }

    public IEnumerable<IOptimizer> NestedOptimizers {
      get {
        if (Algorithm == null) yield break;
        yield return Algorithm;
        foreach (var opt in Algorithm.NestedOptimizers)
          yield return opt;
      }
    }

    private bool IsFinished {
      get {
        var timeHit = ExecutionTime >= MaximumExecutionTime;
        var evalHit = Evaluations >= MaximumEvaluations;
        var targetHit = (Maximization && BestSoFar >= TargetValue || !Maximization && BestSoFar <= TargetValue);

        return timeHit && evalHit && targetHit
          || timeHit && (TerminationCriterium == TerminationCriterium.OnlyByTime
                      || TerminationCriterium == TerminationCriterium.ByTargetAndTime
                      || TerminationCriterium == TerminationCriterium.WhicheverHitsFirst)
          || evalHit && (TerminationCriterium == TerminationCriterium.OnlyByEvaluations
                      || TerminationCriterium == TerminationCriterium.ByTargetAndEvaluations
                      || TerminationCriterium == TerminationCriterium.WhicheverHitsFirst)
          || targetHit && (TerminationCriterium == TerminationCriterium.OnlyByTarget
                        || TerminationCriterium == TerminationCriterium.WhicheverHitsFirst
                        || TerminationCriterium == TerminationCriterium.ByTargetAndTime
                        || TerminationCriterium == TerminationCriterium.ByTargetAndEvaluations);
      }
    }

    [StorableConstructor]
    private IndepdentRandomRestartRun(bool deserializing) : base(deserializing) { }
    private IndepdentRandomRestartRun(IndepdentRandomRestartRun original, Cloner cloner)
      : base(original, cloner) {
      terminationCriterium = original.terminationCriterium;
      maximumExecutionTime = original.maximumExecutionTime;
      maximumEvaluations = original.maximumEvaluations;
      targetValue = original.targetValue;
      executionTime = original.executionTime;
      evaluations = original.evaluations;
      bestSoFar = original.bestSoFar;
      lastAlgorithmExecutionTime = original.lastAlgorithmExecutionTime;
      lastAlgorithmEvaluatedSolutions = original.lastAlgorithmEvaluatedSolutions;
      lastAlgorithmEvaluatedMoves = original.lastAlgorithmEvaluatedMoves;

      algorithm = cloner.Clone(original.algorithm);
      runs = cloner.Clone(original.runs);

      Initialize();
    }
    public IndepdentRandomRestartRun()
      : base() {
      name = ItemName;
      description = ItemDescription;
      terminationCriterium = TerminationCriterium.OnlyByEvaluations;
      maximumExecutionTime = TimeSpan.FromMinutes(1);
      maximumEvaluations = 10000000; // 10 mio
      targetValue = 0;
      executionTime = TimeSpan.Zero;
      evaluations = 0;
      bestSoFar = double.NaN;
      lastAlgorithmExecutionTime = TimeSpan.Zero;
      lastAlgorithmEvaluatedSolutions = 0;
      lastAlgorithmEvaluatedMoves = 0;

      Runs = new RunCollection { OptimizerName = Name };
      Initialize();
    }
    public IndepdentRandomRestartRun(string name)
      : base(name) {
      description = ItemDescription;
      terminationCriterium = TerminationCriterium.OnlyByEvaluations;
      maximumExecutionTime = TimeSpan.FromMinutes(1);
      maximumEvaluations = 10000000; // 10 mio
      targetValue = 0;
      executionTime = TimeSpan.Zero;
      evaluations = 0;
      bestSoFar = double.NaN;
      lastAlgorithmExecutionTime = TimeSpan.Zero;
      lastAlgorithmEvaluatedSolutions = 0;
      lastAlgorithmEvaluatedMoves = 0;

      Runs = new RunCollection { OptimizerName = Name };
      Initialize();
    }
    public IndepdentRandomRestartRun(string name, string description)
      : base(name, description) {
      terminationCriterium = TerminationCriterium.OnlyByEvaluations;
      maximumExecutionTime = TimeSpan.FromMinutes(1);
      maximumEvaluations = 10000000; // 10 mio
      targetValue = 0;
      executionTime = TimeSpan.Zero;
      evaluations = 0;
      bestSoFar = double.NaN;
      lastAlgorithmExecutionTime = TimeSpan.Zero;
      lastAlgorithmEvaluatedSolutions = 0;
      lastAlgorithmEvaluatedMoves = 0;

      Runs = new RunCollection { OptimizerName = Name };
      Initialize();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      if (ExecutionState == ExecutionState.Started) throw new InvalidOperationException(string.Format("Clone not allowed in execution state \"{0}\".", ExecutionState));
      return new IndepdentRandomRestartRun(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      Initialize();
    }

    private void Initialize() {
      if (algorithm != null) RegisterAlgorithmEvents();
    }

    public void Prepare() {
      Prepare(false);
    }
    public void Prepare(bool clearRuns) {
      executionTime = TimeSpan.Zero;
      evaluations = 0;
      lastAlgorithmExecutionTime = TimeSpan.Zero;
      lastAlgorithmEvaluatedSolutions = 0;

      Algorithm.Prepare(clearRuns);
    }
    public void Start() {
      if (ExecutionState == ExecutionState.Prepared) {
        currentRun = new Run(Algorithm) {
          Name = Algorithm.Name + " IRRRun" + Runs.Count
        };
      }
      Algorithm.Start();
    }
    public void Pause() {
      Algorithm.Pause();
    }
    public void Stop() {
      Algorithm.Stop();
    }

    #region Events
    protected override void OnNameChanged() {
      base.OnNameChanged();
      runs.OptimizerName = Name;
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string property) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(property));
    }

    #region IExecutable Events
    public event EventHandler ExecutionStateChanged;
    private void OnExecutionStateChanged() {
      var handler = ExecutionStateChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler ExecutionTimeChanged;
    private void OnExecutionTimeChanged() {
      var handler = ExecutionTimeChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Prepared;
    private void OnPrepared() {
      var handler = Prepared;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Started;
    private void OnStarted() {
      var handler = Started;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Paused;
    private void OnPaused() {
      var handler = Paused;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler Stopped;
    private void OnStopped() {
      var handler = Stopped;
      if (handler != null) handler(this, EventArgs.Empty);
    }
    public event EventHandler<EventArgs<Exception>> ExceptionOccurred;
    private void OnExceptionOccurred(Exception exception) {
      var handler = ExceptionOccurred;
      if (handler != null) handler(this, new EventArgs<Exception>(exception));
    }
    #endregion

    #region Algorithm Events
    private void RegisterAlgorithmEvents() {
      algorithm.ExceptionOccurred += Algorithm_ExceptionOccurred;
      algorithm.ExecutionTimeChanged += Algorithm_ExecutionTimeChanged;
      algorithm.ExecutionStateChanged += Algorithm_ExecutionStateChanged;
      algorithm.Paused += Algorithm_Paused;
      algorithm.Prepared += Algorithm_Prepared;
      algorithm.Started += Algorithm_Started;
      algorithm.Stopped += Algorithm_Stopped;
      algorithm.ProblemChanged += Algorithm_ProblemChanged;
    }
    private void DeregisterAlgorithmEvents() {
      algorithm.ExceptionOccurred -= Algorithm_ExceptionOccurred;
      algorithm.ExecutionTimeChanged -= Algorithm_ExecutionTimeChanged;
      algorithm.ExecutionStateChanged -= Algorithm_ExecutionStateChanged;
      algorithm.Paused -= Algorithm_Paused;
      algorithm.Prepared -= Algorithm_Prepared;
      algorithm.Started -= Algorithm_Started;
      algorithm.Stopped -= Algorithm_Stopped;
      algorithm.ProblemChanged -= Algorithm_ProblemChanged;
    }
    private void Algorithm_ExceptionOccurred(object sender, EventArgs<Exception> e) {
      OnExceptionOccurred(e.Value);
    }
    private void Algorithm_ExecutionTimeChanged(object sender, EventArgs e) {
      ExecutionTime += Algorithm.ExecutionTime - lastAlgorithmExecutionTime;
      lastAlgorithmExecutionTime = Algorithm.ExecutionTime;

      UpdateAlgorithmResults();

      if (IsFinished) {
        Algorithm.Stop();
      }
      OnExecutionTimeChanged();
    }

    private void Algorithm_ExecutionStateChanged(object sender, EventArgs e) {
      OnExecutionStateChanged();
    }
    private void Algorithm_Paused(object sender, EventArgs e) {
      ExecutionTime += Algorithm.ExecutionTime - lastAlgorithmExecutionTime;
      lastAlgorithmExecutionTime = Algorithm.ExecutionTime;

      UpdateAlgorithmResults();
      OnPaused();
    }
    private void Algorithm_Prepared(object sender, EventArgs e) {
      OnPrepared();
    }
    private void Algorithm_Started(object sender, EventArgs e) {
      OnStarted();
    }
    private void Algorithm_Stopped(object sender, EventArgs e) {
      ExecutionTime += Algorithm.ExecutionTime - lastAlgorithmExecutionTime;
      lastAlgorithmExecutionTime = Algorithm.ExecutionTime;

      var bestQuality = UpdateAlgorithmResults();

      foreach (var result in Algorithm.Results) {
        if (result.Name == "QualityPerClock") {
          if (!currentRun.Results.ContainsKey(result.Name))
            currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
          else {
            var dt = (IndexedDataTable<double>)currentRun.Results[result.Name];
            var execTime = ((TimeSpanValue)currentRun.Results["ExecutionTime"]).Value.TotalSeconds;
            var best = dt.Rows.First().Values.Last().Item2;
            var resultDt = (IndexedDataTable<double>)result.Value;
            foreach (var tupl in resultDt.Rows.First().Values) {
              if (Maximization && tupl.Item2 > best || !Maximization && tupl.Item2 < best) {
                dt.Rows.First().Values.Add(Tuple.Create(execTime + tupl.Item1, tupl.Item2));
                best = tupl.Item2;
              }
            }
          }
        } else if (result.Name == "QualityPerEvaluations") {
          if (!currentRun.Results.ContainsKey(result.Name))
            currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
          else {
            var dt = (IndexedDataTable<double>)currentRun.Results[result.Name];
            var evalSols = ((DoubleValue)currentRun.Results["EvaluatedSolutions"]).Value;
            var evalMoves = ((DoubleValue)currentRun.Results["EvaluatedMoves"]).Value;
            var best = dt.Rows.First().Values.Last().Item2;
            var resultDt = (IndexedDataTable<double>)result.Value;
            foreach (var tupl in resultDt.Rows.First().Values) {
              if (Maximization && tupl.Item2 > best || !Maximization && tupl.Item2 < best) {
                dt.Rows.First().Values.Add(Tuple.Create(evalSols + moveCostPerSolution * evalMoves + tupl.Item1, tupl.Item2));
                best = tupl.Item2;
              }
            }
          }
        } else if (result.Name == "EvaluatedSolutions") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name == "EvaluatedMoves") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name == "ExecutionTime") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name == "BestQuality") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name.ToLower().EndsWith("solution") && BestSoFar == bestQuality) {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        }
      }
      foreach (var result in Algorithm.Results) {
        if (result.Name == "QualityPerClock") {

        } else if (result.Name == "QualityPerEvaluations") {
        } else if (result.Name == "EvaluatedSolutions") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name == "EvaluatedMoves") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name == "ExecutionTime") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name == "BestQuality") {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        } else if (result.Name.ToLower().EndsWith("solution") && BestSoFar == bestQuality) {
          currentRun.Results.Add(result.Name, (IItem)result.Value.Clone());
        }
      }


      if (IsFinished) {

      }

      // TODO
      var cloner = new Cloner();
      var algRun = cloner.Clone(Algorithm.Runs.Last());
      Runs.Add(algRun);
      Algorithm.Runs.Clear();
      OnStopped();
    }

    private double UpdateAlgorithmResults() {
      IResult evaluationsResult;
      if (Algorithm.Results.TryGetValue("EvaluatedSolutions", out evaluationsResult)) {
        var evals = ((IntValue)evaluationsResult.Value).Value;
        Evaluations += evals - lastAlgorithmEvaluatedSolutions;
        lastAlgorithmEvaluatedSolutions = evals;
      }
      if (Algorithm.Results.TryGetValue("EvaluatedMoves", out evaluationsResult)) {
        var evals = ((IntValue)evaluationsResult.Value).Value;
        Evaluations += moveCostPerSolution * (evals - lastAlgorithmEvaluatedMoves);
        lastAlgorithmEvaluatedMoves = evals;
      }
      if (Algorithm.Results.TryGetValue("BestQuality", out evaluationsResult)) {
        var bestQuality = ((DoubleValue)evaluationsResult).Value;
        if (double.IsNaN(BestSoFar)
            || Maximization && bestQuality > BestSoFar
            || !Maximization && bestQuality < BestSoFar)
          BestSoFar = bestQuality;
        return bestQuality;
      }
      return double.NaN;
    }

    private void Algorithm_ProblemChanged(object sender, EventArgs e) {
      var soProblem = Algorithm.Problem as ISingleObjectiveHeuristicOptimizationProblem;
      if (soProblem == null) return;
      var maxParam = soProblem.MaximizationParameter as IValueParameter<BoolValue>;
      if (maxParam != null)
        Maximization = maxParam.Value.Value;
      var bkParam = soProblem.BestKnownQualityParameter as IValueParameter<DoubleValue>;
      if (bkParam != null && bkParam.Value != null)
        TargetValue = bkParam.Value.Value;
    }
    #endregion
    #endregion
  }
}
